# -*- coding: iso-8859-1 -*-
"""Vergleich von Gauss- und Binomialverteilung

Vgl. Skript, Folie "Hufig auftretende Verteilungen"
"""


import Tkinter, math, operator

class PlottableFunction:
	"""Diese Klasse reprsentiert ein Funktion, die auf einem canvas
	geplottet werden kann.  (Langweiliger GUI-Kram)
	"""
	def __init__(self, func, canvas):
		self._func = func

	def getVals(self, xRange, numPoints, step):
		discWidth = (xRange[1]-xRange[0])/float(numPoints)
		return map(self._func, [(x+0.5)*discWidth+xRange[0]
			for x in range(0, numPoints, step)])


class TwoFunPlotter(Tkinter.Tk):
	"""Diese Funktion ist ein Toplevel-Fenster, das einen Canvas
	mit zwei Plottable Functions enthlt. (Langweiler GUI-Kram)
	"""
	def __init__(self, mFunc1, mFunc2, *args, **kwargs):
		apply(Tkinter.Tk.__init__, (self,)+args[2:], kwargs)
		self._makeCanvas()
		self._pFunc1 = PlottableFunction(mFunc1, self._funCanvas)
		self._pFunc2 = PlottableFunction(mFunc2, self._funCanvas)
		self.bind("d", self._drawStuff)
		self.bind("q", lambda ev, self=self: self.quit())
		self.plotRange = (0,10)
		self._id1 = None
		self._id2 = None

	def _makeCanvas(self):
		self._funCanvas = Tkinter.Canvas(self)
		self._funCanvas.pack(expand=1, fill=Tkinter.BOTH)
	
	def _clearFuns(self):
		if not self._id1 is None:
			self._funCanvas.delete(self._id1)
		if not self._id2 is None:
			self._funCanvas.delete(self._id2)

	def _getGeometry(self):
		self._xSize = self.winfo_width()
		self._ySize = self.winfo_height()-2

	def _plotStuff(self, vals, fMin, fMax, step):
		return self._funCanvas.create_line(
			zip(range(0, self._xSize, step), 
				[int(((fMax-y)/float(fMax-fMin)-fMin)*self._ySize)+1 
				for y in vals]))

	def _drawStuff(self, ev=None):
		try:
			step = 3
			self._clearFuns()
			self._getGeometry()
			vals1 = self._pFunc1.getVals(self.plotRange, self._xSize, step)
			vals2 = self._pFunc2.getVals(self.plotRange, self._xSize, step)
			minF = min(min(vals1), min(vals2))
			maxF = max(max(vals1), max(vals2))
			self._id1 = self._plotStuff(vals1, minF, maxF, step)
			self._id2 = self._plotStuff(vals2, minF, maxF, step)
		except Tkinter.TclError:
			self._id1 = None
			self._id2 = None


def faculty(n):
	"""gibt die Fakultt von n zurck (und sollte fr groe n auf
	eine Nherung umstellen.)
	"""
	return reduce(operator.mul, range(1, n+1), 1)


def binomial(up, down):
	"""gibt den Binomialkoeffizienten up ber down zurck.	Auch hier
	sollte fr grere Zahlen genhert werden, z.B. per Stirling.
	"""
	return faculty(up)/(faculty(down)*faculty(up-down))


class FunctionWithParameters:
	"""Abstrakte Basisklasse fr Funktionen (sorgt eigentlich nur dafr,
	dass die unten definierten Klassen nicht mit __call__ rummachen
	mssen und ist von fraglichem Wert).
	"""
	def __call__(self, arg):
		return self._func(arg)


class GaussFun(FunctionWithParameters):
	"""Diese Klasse reprsentiert eine Gauss-Funktion -- sie
	hat die beiden Parameter mu (Erwartungswert) und sigma
	(Standardabweichung), die hier ffentliche Attribute zum
	Lesen und Schreiben sind.
	"""
	def __init__(self):
		self.mu = 0
		self.sigma = 1
	
	def _func(self, arg):
		return 1/math.sqrt(2*3.14159)/self.sigma*\
			math.exp(-(arg-self.mu)**2/2/self.sigma**2)


class BernoulliFun(FunctionWithParameters):
	"""Diese Klasse reprsentiert eine Bernoulli-Funktion mit den
	Parametern p (Erfolgswahrscheinlichkeit) und numBalls (Zahl
	der Experimente), die auch hier ffentliche Attribute zum
	Lesen und Schreiben sind.
	"""
	def __init__(self):
		self.p = 0.5
		self.numBalls = 20

	def _func(self, arg):
		arg = int(round(arg))
		if arg<0 or arg>self.numBalls:
			return 0
		return binomial(self.numBalls, arg)*self.p**arg*\
			(1-self.p)**(self.numBalls-arg)


class GaussVsBernComparer(TwoFunPlotter):
	"""Diese Klasse spezialisiert den TwoFunPlotter auf
	eine Gauss- und eine Bernoulli-Verteilung und sorgt
	fr die "BenutzerInnenschnittstelle".  Interessant ist
	der _changePars-Callback und die von ihm benutzte
	setParameters-Methode.
	"""
	def __init__(self, *args, **kwargs):
		self.gaussF = GaussFun()
		self.bernF = BernoulliFun()
		apply(TwoFunPlotter.__init__, (self, self.gaussF, self.bernF)+args,
			kwargs)
		self.numBalls = 10
		self._dispLabel = Tkinter.Label(self, text="p=0.5")
		self._labelId = self._funCanvas.create_window(
			50, 50, window=self._dispLabel)
		self.zoom = 1
		self.p = 0.5
		self.bind('8', self._changePars)
		self.bind('2', self._changePars)
		self.bind('4', self._changePars)
		self.bind('6', self._changePars)
		self.bind('z', self._changeZoom)
		self.bind('Z', self._changeZoom)
		self.bind('<Configure>', self._resize)

	def _changePars(self, ev):
		if ev.char == '8':
			self.numBalls += 1
		if ev.char == '2' and self.numBalls>2:
			self.numBalls -= 1
		if ev.char == '4':
			if self.p>0.02:
				self.p -= 0.02
			else:
				if self.p>1e-4:
					self.p -= 1e-3
		if ev.char == '6' and self.p<0.98:
			self.p += 0.02
		self.setParameters()

	def _changeZoom(self, ev):
		if ev.char == 'Z':
			if self.zoom>1.1:
				self.zoom /= 1.5
		if ev.char == 'z':
			if self.zoom<100:
				self.zoom *= 1.5
		self.setParameters()

	def _resize(self, ev=None):
		self._getGeometry()
		self._funCanvas.coords(self._labelId, self._xSize/2-40,
			self._ySize-20)
		self.setParameters()

	def setParameters(self):
		"""sorgt dafr, dass die Gauss-Verteilung die jeweils beste
		Approximation fr die gewhlte Bernoulli-Verteilung ist.
		"""
		self.bernF.numBalls = self.numBalls
		self.bernF.p = self.p
		self.gaussF.mu = self.numBalls*self.p
		self.gaussF.sigma = math.sqrt(self.numBalls*self.p*(1-self.p))
		self.plotRange = (-1, int((self.numBalls+1)/self.zoom))
		self._dispLabel.configure(text="p=%0.3f N=%d"%(self.p, self.numBalls))
		self._drawStuff()


if __name__=="__main__":
	p = GaussVsBernComparer(GaussFun(), BernoulliFun())
	p.setParameters()
	p.geometry("600x400")
	p.mainloop()
